local addonName, addon = ...

local format, header, playerClass, select = string.format, addon.header, addon.playerClass, select
local Execute, SetMouseoverSensitivity = header.Execute, addon.SetMouseoverSensitivity

local actionDescAppend, actionMode, actionStatusOverride, actionTag, initialized = { }, { }, { }, { }

local clearTag = { custom = true, gtaoe = true, harm = true, help = true, ["harm/help"] = true, override = true, rez = true, substitution = true }
local overrideDefaults = { enable = true, mode = false, useCustomMode = false, customMode = false, status = false, useCustomParameters = false, customParameters = false }
local tagImpliesClassAbility = { pet = true, pvp = true, spec = true, talent = true }
local validActionMode = { harm = true, help = true, ["harm/help"] = true, rez = true, gtaoe = true }
local validActionType = { item = true, spell = true }

local ERROR_TYPE = "bad argument #%d to %q (%s expected, got %s)"
local ERROR_VALUE = "bad argument #%d to %q (invalid value %q)"

for mode in pairs(validActionMode) do
	Execute(header, format("validActionMode[%q] = true", mode))
end

addon.actionTag = setmetatable(actionTag, { __index = function(self, tagID)
	self[tagID] = { }
	return self[tagID]
end })

local function SetActionInfo(actionType, typeID, mode, statusOverride, ...)
	if statusOverride == "npc" then
		if mode == 'harm' or mode == 'help' then
			statusOverride = format("%s=%s", mode, "noharm,nohelp,exists")
		else
			statusOverride = nil
		end
	end

	local actionID = format("%s:%s", actionType, typeID)
	actionMode[actionID], actionStatusOverride[actionID] = mode, statusOverride

	local descAppend
	for index = 1, select('#', ...) do
		local tag = select(index, ...)
		local param, args = tag:match("^([^:]+):([^:]+)$")
		if param then
			if descAppend then
				descAppend = format("%s;%s", descAppend, tag)
			else
				descAppend = tag
			end
			if param == 'spec' then
				for arg in args:gmatch("[^/]+") do
					local tag = format("%s:%s", 'spec', arg)
					actionTag[tag][actionID] = tag
				end
			elseif param == 'talent' then
				local tier, column, spec = args:match("(%d)/(%d)/(%d)")
				if spec then
					local tag = format("%s:%s", 'spec', spec)
					actionTag[tag][actionID] = tag
				end
			elseif param == "petSpec" then
				param = 'pet'
			end
			tag = param
		end
		if tag ~= 'spec' then
			actionTag[tag][actionID] = tag
		end
		if tagImpliesClassAbility[tag] then
			actionTag[playerClass][actionID] = playerClass
		end
	end
	actionDescAppend[actionID] = descAppend
	actionTag[actionType][actionID] = actionType
end

local ApplyActionOverride
do
	local snippetClearOverride = ([[
		actionMode[%1$q] = nil
		actionStatusOverride[%1$q] = nil
		parameters.unitframe[%1$q] = nil
		parameters.unitmodel[%1$q] = nil
	]]):trim():gsub("%s+", " ")

	local ApplyActionParameters = addon.ApplyActionParameters

	function ApplyActionOverride(actionID, override)
		Execute(header, format(snippetClearOverride, actionID))
		SetMouseoverSensitivity(actionID, false)
		for tag in pairs(clearTag) do
			actionTag[tag][actionID] = nil
		end

		local mode
		if override.enable then
			if override.useCustomParameters then
				if override.customParameters then
					ApplyActionParameters(actionID, override.customParameters)
					mode = "custom"
				else
					-- Shouldn't be able to get here but checking just in case
				end
			else
				if not override.useCustomMode then
					mode = override.mode or actionMode[actionID]
				else
					mode = override.customMode or actionMode[actionID]
				end
				if mode then
					if not validActionMode[mode] then
						ApplyActionParameters(actionID, mode)
						mode = "custom"
					end
					Execute(header, format("actionMode[%q] = %q", actionID, mode))
					local statusOverride = override.status
					if statusOverride then
						actionTag["custom"][actionID] = true
					else
						statusOverride = actionStatusOverride[actionID]
					end
					if statusOverride then
						Execute(header, format("actionStatusOverride[%q] = %q", actionID, statusOverride))
						actionTag["substitution"][actionID] = true
					end
				end
			end
		else
			Execute(header, format("actionMode[%q] = %q", actionID, "ignore"))
			if override.useCustomParameters then
				if override.customParameters then									-- Error if this fails?
					mode = "custom"
				else
					-- Shouldn't be able to get here but checking just in case
				end
			else
				if not override.useCustomMode then
					mode = override.mode or actionMode[actionID]
				else
					mode = override.customMode or actionMode[actionID]
				end
				if mode then
					if not validActionMode[mode] then
						mode = "custom"
					end
					if override.status then
						actionTag["custom"][actionID] = true
						actionTag["substitution"][actionID] = true
					elseif actionStatusOverride[actionID] then
						actionTag["substitution"][actionID] = true
					end
				end
			end
		end
		if mode then
			actionTag[mode][actionID] = mode
		end
		for field, default in pairs(overrideDefaults) do
			if (override[field] or false) ~= default then
				actionTag["override"][actionID] = true
				break
			end
		end
	end

	function addon.ApplyActionOverride(actionID, override)
		addon.SafeCall(ApplyActionOverride, actionID, override)
	end
end

local function ApplyActionOverrides()
	local override, overrideDefaults = addon.db.profile.override, addon.defaultSettings.override['**']
	local rawget = rawget
	for actionID in pairs(actionMode) do
		if not rawget(override, actionID) then
			ApplyActionOverride(actionID, overrideDefaults)						-- Avoid creating table
		end
	end
	local playerClass = addon.playerClass
	for actionID, override in pairs(override) do
		if not override.class or override.class == playerClass then
			ApplyActionOverride(actionID, override)
		end
	end
end

function addon.ApplyActionOverrides()
	addon.ApplyActionOverrides = ApplyActionOverrides
	ApplyActionOverrides()
	initialized = true
end

--[[----------------------------------------------------------------------------
Public API
------------------------------------------------------------------------------]]
local function SoftError(message)
	local ok, err = pcall(error, message, 4)
	geterrorhandler()(err)
end

_G[addonName].SetActionInfo = function(actionType, typeID, mode, statusOverride, ...)
	if type(actionType) ~= 'string' then
		return SoftError(format(ERROR_TYPE, 1, addonName .. ".SetActionInfo", 'string', type(actionType)))
	elseif not validActionType[actionType] then
		return SoftError(format(ERROR_VALUE, 1, addonName .. ".SetActionInfo", actionType))
	elseif type(typeID) ~= 'number' then
		return SoftError(format(ERROR_TYPE, 2, addonName .. ".SetActionInfo", 'number', type(typeID)))
	elseif type(mode) ~= 'string' then
		return SoftError(format(ERROR_TYPE, 3, addonName .. ".SetActionInfo", 'string', type(mode)))
	elseif not (validActionMode[mode] or validActionMode[SecureCmdOptionParse(mode)]) then
		return SoftError(format(ERROR_VALUE, 3, addonName .. ".SetActionInfo", mode))
	elseif statusOverride ~= nil and type(statusOverride) ~= 'string' then
		return SoftError(format(ERROR_TYPE, 4, addonName .. ".SetActionInfo", "string or nil", type(statusOverride)))
	else
		for index = 1, select('#', ...) do
			local tag_type = type(select(index, ...))
			if tag_type ~= 'string' and tag_type ~= 'nil' then
				return SoftError(format(ERROR_TYPE, 4 + index, addonName .. ".SetActionInfo", "string or nil", tag_type))
			end
		end
	end

	SetActionInfo(actionType, typeID, mode, statusOverride, ...)
	if initialized then
		local actionID = format("%s:%s", actionType, typeID)
		addon.ApplyActionOverride(actionID, addon.db.profile.override[actionID])
		if addon.config then
			addon.config[1]:Refresh()
		end
	end
end

--[[----------------------------------------------------------------------------
Initialize
------------------------------------------------------------------------------]]
addon.actionMode = actionMode
addon.actionDescAppend, addon.actionStatusOverride = actionDescAppend, actionStatusOverride
addon.overrideDefaults = overrideDefaults
addon.validActionMode, addon.validActionType = validActionMode, validActionType
addon.SetActionInfo = SetActionInfo
