local addonName, addon, L = LibStub('LibOptionsAssist-1.1').FetchData()

local actionMode, actionName, actionTag = addon.actionMode, addon.actionName, addon.actionTag
local actionDescAppend, actionStatusOverride = addon.actionDescAppend, addon.actionStatusOverride
local classTag, validActionMode = actionTag[addon.playerClass], addon.validActionMode

local format, gsub, match, trim = string.format, string.gsub, string.match, string.trim
local next, strsplit = next, strsplit
local GetActionInfo = GetActionInfo
local IsHarmfulItem, IsHelpfulItem = IsHarmfulItem, IsHelpfulItem
local IsHarmfulSpell, IsHelpfulSpell = IsHarmfulSpell, IsHelpfulSpell

local AdjustParameters, GetModeParameters = addon.AdjustParameters, addon.GetModeParameters

local modeValue = { [1] = 'harm', [2] = 'help', [3] = "harm/help", [4] = "rez", [5] = "gtaoe",
					['harm'] = 1, ['help'] = 2, ["harm/help"] = 3, ["rez"] = 4, ["gtaoe"] = 5 }

local _override, _settings, overrideTab

--[[----------------------------------------------------------------------------
Tag Support
------------------------------------------------------------------------------]]
local className, playerClass = UnitClass('player')

local tagName = {
	['pet']			= PET or "Pet",
	['pvp']			= PVP or "PvP",
	[playerClass]	= className
}

for index = 1, GetNumSpecializations() do										-- Player specializations
	local _, name = GetSpecializationInfo(index)
	tagName[format("%s:%s", 'spec', index)] = name
end

for index = 1, GetNumSpecializations(false, true) do							-- Pet specializations
	local _, name = GetSpecializationInfo(index, false, true)
	tagName[format("%s:%s", "petSpec", index)] = name
end

--[[----------------------------------------------------------------------------
General Support
------------------------------------------------------------------------------]]
local UpdateActionNames, UpdateAvailableFilters, UpdateOverrideEntries			-- Defined later

local function GetActionInfo(info)
	local actionID = info[#info - 1]
	local actionType, typeID = strsplit(":", actionID)
	return actionID, actionType, tonumber(typeID) or 0
end

local function GetModeFromAPI(info)
	local value, IsHarmful, IsHelpful
	local actionID, actionType, typeID = GetActionInfo(info)
	if actionType == 'spell' then
		IsHarmful, IsHelpful, typeID = IsHarmfulSpell, IsHelpfulSpell, actionName[actionID]
	else
		IsHarmful, IsHelpful = IsHarmfulItem, IsHelpfulItem
	end
	if IsHarmful(typeID) then
		value = not IsHelpful(typeID) and 'harm' or "harm/help"
	else
		value = IsHelpful(typeID) and 'help'
	end
	return value
end

local function GetOverride(info)
	local size = #info
	return _override[info[size - 1]][info[size]]
end

local function GetOverrideEntryDesc(info)
	return format("|cff808080%s|r", info[#info])
end

local function GetOverrideFilter(info)
	return _settings.overrideFilter[info[#info]]
end

local function GetCustomMode(info)
	local actionID = info[#info - 1]
	local value = _override[actionID].mode or actionMode[actionID]
	if validActionMode[value] then
		value = nil
	end
	return value
end

local function GetOverrideMode(info)
	local actionID = info[#info - 1]
	local value = _override[actionID].mode
	if not value then
		value = actionMode[actionID]
		if not validActionMode[value] then
			value = GetModeFromAPI(info)
		end
	end
	return value
end

local function GetOverrideModeDesc(info)
	local api, hardCoded = GetModeFromAPI(info), actionMode[info[#info - 1]]
	hardCoded = validActionMode[hardCoded] and hardCoded
	return format(L["Select the mode for this action.%s%s%s"], (api or hardCoded) and "|cff808080\n" or "", hardCoded and format("\nHard coded: %s", hardCoded) or "", api and format("\nAPI default: %s", api) or "")
end

-- FIX: Allow versatility with spacing: "harm,nodead=dead" vs "harm, nodead=dead" vs "harm,%s*nodead=dead"
local function GetOverrideParameters(info)
	local actionID = info[#info - 1]
	local value = _override[actionID].customParameters
	if not value then
		local mode = GetOverrideMode(info)
		if mode then
			value = GetModeParameters(mode, _settings, nil, true, true)

			local statusOverride = _override[actionID].status or actionStatusOverride[actionID]
			if statusOverride then
				value = gsub(value, "([%[,])%s*(%S+)%s*([%],])", "%1%2%3")
				for substitution in statusOverride:gmatch("[^;]+") do
					local oldStatus, newStatus = match(substitution, "^([^=]+)=([^=]+)$")
					if newStatus then
						value = gsub(value, format("([%%[,])%%s*%s%%s*([%%],])", oldStatus), format("%%1%s%%2", newStatus))
					end
				end
				value = AdjustParameters(value)
			end

		end
	end
	return value or ""
end

local function IsOverrideDisabled(info)
	local field, override = info[#info], _override[info[#info - 1]]
	if field ~= "customParameters" then
		if field ~= "customMode" then
			return override.useCustomParameters
		end
		return override.useCustomParameters or override.mode ~= "custom"
	end
	return not override.useCustomParameters
end

local function ProcessAllTags(info)
	local filter, value = _settings.overrideFilter, info[#info] == "CheckAllTags"
	for tag in pairs(actionTag) do
		filter[tag] = value
	end
end

local function SetOverride(info, value)
	local size = #info
	local actionID, field = info[size - 1], info[size]
	_override[actionID][field] = value
	_override[actionID].class = classTag[actionID]
	addon.ApplyActionOverride(actionID, _override[actionID])
end

local function SetOverrideFilter(info, value)
	_settings.overrideFilter[info[#info]] = value
end

local function UseCustomMode(info)
	local value = _override[info[#info - 1]].useCustomMode
	if value == nil then
		value = GetCustomMode(info)
	end
	return value
end

local overrideEntryArgs = {
	enable = {
		type = 'toggle', width = 0.75,
		order = 1,
		name = L["Enable"],
		desc = L["Determines if %s will modify the behavior of this action."]:format(addonName)
	},
	mode = {
		type = 'select',
		order = 2,
		name = L["Mode"],
		desc = GetOverrideModeDesc,
		disabled = function(info)
			return _override[info[#info - 1]].useCustomParameters or UseCustomMode(info)
		end,
		get = function(info)
			return modeValue[GetOverrideMode(info)]
		end,
		set = function(info, value)
			value = modeValue[value]
			if value == actionMode[info[#info - 1]] then
				value = nil
			end
			SetOverride(info, value)
		end,
		values = { unpack(modeValue) }
	},
	useCustomMode = {
		type = 'toggle', width = 'double',
		order = 3,
		name = L["Custom Mode"],
		desc = L["Allow custom parameters to determine the mode for this action."],
		disabled = function(info)
			return _override[info[#info - 1]].useCustomParameters
		end,
		get = UseCustomMode,
		set = function(info, value)
			if value and GetCustomMode(info) then
				value = nil
			end
			SetOverride(info, value)
		end
	},
	customMode = {
		type = 'input', width = 'full',
		order = 4,
		name = L["Mode Parameters"],
		desc = L["Enter macro-style parameters that will determine the mode used for this action.\n\n|cff808080Ex: [spec:1,talent:1/1] harm/help; harm|r"],
		disabled = function(info)
			return not UseCustomMode(info)
		end,
		get = function(info)
			return GetCustomMode(info) or ""
		end,
		set = function(info, value)
			local actionID = info[#info - 1]
			value = AdjustParameters(value)
			local value_lower = value:lower()
			if validActionMode[value_lower] then
				_override[actionID].mode = value_lower ~= actionMode[actionID] and value_lower or nil
				_override[actionID].useCustomMode = nil
				value = nil
			elseif value == actionMode[actionID] then
				value = nil
			end
			SetOverride(info, value)
		end
	},
	status = {
		type = 'input', width = 'full',
		order = 5,
		name = L["Condition Substitution"],
		desc = L["Enter semicolon delimited condition substitutions for this action's targeting parameters.\n\nEx: dead=exists;help=noharm"],
		disabled = function(info)
			return _override[info[#info - 1]].useCustomParameters
		end,
		get = function(info)
			return GetOverride(info) or actionStatusOverride[info[#info - 1]] or ""
		end,
		set = function(info, value)
			value = gsub(value, "%s+", "")
			if value == (actionStatusOverride[info[#info - 1]] or "") then
				value = nil
			end
			SetOverride(info, value)
		end
	},
	useCustomParameters = {
		type = 'toggle', width = 'double',
		order = -4,
		name = L["Custom Targeting"],
		desc = L["Allow custom parameters to determine the targeting behavior of this action."],
		set = function(info, value)
			local actionID = info[#info - 1]
			local parameters = GetOverrideParameters(info)
			if value then
				if not _override[actionID].customParameters then
					_override[actionID].customParameters = parameters
				end
			elseif _override[actionID].customParameters == parameters then
				_override[actionID].customParameters = nil
			end
			SetOverride(info, value)
		end
	},
	customParameters = {
		type = 'input', multiline = 3, width = 'full',
		order = -3,
		name = L["Targeting Parameters"],
		desc = L["Enter macro-style parameters that will determine the targeting behavior of this action."],
		disabled = function(info)
			return not _override[info[#info - 1]].useCustomParameters
		end,
		get = GetOverrideParameters,
		set = function(info, value)
			if match(value, "%S") then
				value = AdjustParameters(value)
			else
				value = nil
			end
			SetOverride(info, value)
		end
	},
	tags = {
		type = 'description', width = 'full',
		order = -2
	},
	reset = {
		type = 'execute',
		order = -1,
		name = L["Reset"],
		desc = L["Reset these overrides to their defaults."],
		disabled = function(info)
			return not actionTag["override"][info[#info - 1]]
		end,
		func = function(info)
			local actionID = info[#info - 1]
			local override = _override[actionID]
			for field, default in pairs(addon.overrideDefaults) do
				override[field] = default or nil
			end
			override.class = classTag[actionID]
			addon.ApplyActionOverride(actionID, override)
			GameTooltip:Hide()													-- Fix for AceGUI bug
		end
	}
}

do
	local tags = { }

	local sortableNames = setmetatable({ }, {
		__index = function(self, index)
			self[index] = index:lower()
			return self[index]
		end
	})

	local function SortFunction(a, b)
		return sortableNames[a] < sortableNames[b]
	end

	overrideEntryArgs.tags.name = function(info)
		local actionID, numTags = info[#info - 1], 0
		wipe(tags)
		for tag, actions in pairs(actionTag) do
			if actions[actionID] then
				numTags = numTags + 1
				tags[numTags] = tagName[tag] or tag
			end
		end
		if numTags > 1 then
			table.sort(tags, SortFunction)
		end
		return format("\n|cffffd100Tags:|r %s\n", strjoin(", ", unpack(tags)))
	end
end

local overrideOptions = {
	type = 'group', childGroups = 'tab',
	onTabSelect = function(info, tab)
		if tab == overrideTab then return end
		overrideTab = tab
		if tab == "actions" then
			UpdateOverrideEntries()
		elseif tab == "filter" then
			UpdateAvailableFilters()
		end
	end,
	args = {
		actions = {
			type = 'group',
			order = 1,
			name = L["Actions"],
			desc = L["Options for changing the behavior of individual actions."],
			get = GetOverride,
			set = SetOverride,
			args = { }
		},
		filter = {
			type = 'group',
			order = 2,
			name = L["Filter"],
			desc = L["Options for determining which actions should be displayed under the Actions tab."],
			get = GetOverrideFilter,
			set = SetOverrideFilter,
			args = {
				filterMode = {
					type = 'select',
					order = 1,
					name = L["Mode"],
					desc = L["Determines how the selected tags will be used to filter displayed actions:\n\nExclusive = Match all\n\nInclusive = Match any"],
					get = addon._Get,
					set = addon._Set,
					values = { L["Exclusive"], L["Inclusive"] }
				},
				PreTagsBuffer = {
					type = 'description', width = 'full',
					order = 2,
					name = " "
				},
				PostTagsBuffer = {
					type = 'description', width = 'full',
					order = -3,
					name = " "
				},
				CheckAllTags = {
					type = 'execute',
					order = -2,
					name = L["Check All"],
					desc = L["Check every tag listed."],
					func = ProcessAllTags
				},
				UncheckAllTags = {
					type = 'execute',
					order = -1,
					name = L["Uncheck All"],
					desc = L["Uncheck every tag listed."],
					func = ProcessAllTags
				}
			}
		}
	}
}

do
	local tagElements = setmetatable({ }, {
		__index = function(self, key)
			self[key] = {
				type = 'toggle',
				name = tagName[key] or key
			}
			return self[key]
		end
	})

	function UpdateAvailableFilters()
		local filter, filterArgs = _settings.overrideFilter, overrideOptions.args.filter.args

		for tag in pairs(tagElements) do
			filterArgs[tag] = nil
		end

		for tag, actions in pairs(actionTag) do
			if next(actions) then
				filterArgs[tag] = tagElements[tag]
			else
				filter[tag] = false
			end
		end
	end
end

do
	local filteredIDs, nameCount, numSpecs = { }, { }, GetNumSpecializations()
	local tagPriority = { "(talent):([^:]+)", "(spell:%d+)", "(item:%d+)", "(pet:%a+)", "(petSpec:%d)", "(pet)", "(spec):([^:]+)", "(quest):(%d+)" }

	local questTitle
	do
		local GetTitleForQuestID, RequestLoadQuestByID = C_QuestLog.GetTitleForQuestID, C_QuestLog.RequestLoadQuestByID
		local queue, registered = { }

		local function OnQuestDataLoad(self, event, questID, success)
			if queue[questID] then
				queue[questID] = nil
				if not next(queue) then
					registered = nil
					addon:UnregisterEvent(event)
				end
				if success then
					local title = GetTitleForQuestID(questID)
					if title then
						questTitle[questID] = title
						UpdateActionNames()
					end
				end
			end
		end

		questTitle = setmetatable({ }, { __index = function(self, questID)
			if HaveQuestData(questID) then
				local title = GetTitleForQuestID(questID)
				if title then
					self[questID] = title
					return title
				end
			end
			if not queue[questID] then
				queue[questID] = true
				if not registered then
					registered = true
					addon:RegisterEvent('QUEST_DATA_LOAD_RESULT', OnQuestDataLoad)
				end
				RequestLoadQuestByID(questID)
			end
		end })
	end

	local function GetModifiedActionName(actionID)
		local action_name = actionName[actionID]
		if (nameCount[action_name] or 0) > 1 then
			local append = actionDescAppend[actionID]
			if append then
				local capture, args
				for index = 1, #tagPriority do
					capture, args = append:match(tagPriority[index])
					if capture then break end
				end
				if capture then
					append = nil
					if capture == 'talent' then
						local tier, column, spec = args:match("(%d)/(%d)/(%d)")
						if tier then
							tier, column, spec = tonumber(tier), tonumber(column), tonumber(spec)
							if tier and column and spec then
								local _, talentName = GetTalentInfoBySpecialization(spec, tier, column)

								if action_name ~= talentName then

									local needSpecName
									for index = 1, numSpecs do
										if index ~= spec then
											local _, name = GetTalentInfoBySpecialization(index, tier, column)
											if name == talentName then
												needSpecName = true
												break
											end
										end
									end

									if needSpecName then
										local tag = format("%s:%s", 'spec', spec)
										append = format("%s, %s", tagName[tag], talentName)
									else
										append = talentName
									end

								end
							end
						end
					elseif capture == 'spec' then
						for index in args:gmatch("[^/]+") do
							index = tonumber(index) or 0
							if index >= 1 and index <= numSpecs then
								local tag = format("%s:%s", 'spec', index)
								if append then
									append = format("%s, %s", append, tagName[tag])
								else
									append = tagName[tag]
								end
							end
						end
					elseif capture == "quest" then
						local questID = tonumber(args)
						if questID then
							append = questTitle[questID]
						else
							append = nil
						end
					else
						append = tagName[capture] or actionName[capture]-- or capture
					end
				else
					append = nil
				end
			end
			if not append then
				local canAppend
				for index = 1, numSpecs do
					local specTag = format("%s:%s", 'spec', index)
					if actionTag[specTag][actionID] then
						if append then
							append = format("%s, %s", append, tagName[specTag])
						else
							append = tagName[specTag]
						end
					else
						canAppend = true
					end
				end
				if not canAppend then
					append = nil
				end
			end
			if actionTag['pvp'][actionID] then
				if append then
					append = format("%s, %s", append, tagName['pvp'])
				else
					append = tagName['pvp']
				end
			end
			if append then
				action_name = format("%s (%s)", action_name, append)
			end
		end
		return action_name or actionID
	end

	local actionElements = setmetatable({ }, {
		__index = function(self, actionID)
			self[actionID] = {
				type = 'group',
				name = GetModifiedActionName(actionID),
				desc = GetOverrideEntryDesc,
				args = overrideEntryArgs
			}
			return self[actionID]
		end
	})

	local emptyElement = {
		type = 'description',
		name = ("\n\n        %s"):format(L["No matches found"])
	}

	function UpdateOverrideEntries()
		local filter, actionsArgs = _settings.overrideFilter, overrideOptions.args.actions.args

		wipe(filteredIDs)
		if _settings.filterMode == 2 then										-- Inclusive (OR)
			for tag, actions in pairs(actionTag) do
				if filter[tag] then
					for actionID in pairs(actions) do
						filteredIDs[actionID] = true
					end
				end
			end
		else																	-- Exclusive (AND)
			local exclude
			for tag, actions in pairs(actionTag) do
				if filter[tag] and next(actions) then
					if exclude then
						for actionID in pairs(filteredIDs) do
							if not actions[actionID] then
								filteredIDs[actionID] = nil
							end
						end
					else
						for actionID in pairs(actions) do
							filteredIDs[actionID] = true
						end
						exclude = true
					end
				end
			end
		end

		wipe(actionsArgs)
		if next(filteredIDs) then
			for actionID in pairs(filteredIDs) do
				actionsArgs[actionID] = actionElements[actionID]
			end
		else
			actionsArgs["empty"] = emptyElement
		end
	end

	do
		local queued

		local function QueuedUpdateActionNames()
			queued = nil

			wipe(nameCount)
			for _, name in pairs(actionName) do
				nameCount[name] = (nameCount[name] or 0) + 1
			end
			for actionID, element in pairs(actionElements) do
				element.name = GetModifiedActionName(actionID)
			end

			if overrideTab == "actions" then
				addon.config[1]:Refresh()
			end
		end

		function UpdateActionNames()
			if queued then return end
			queued = true
			C_Timer.After(0, QueuedUpdateActionNames)
		end

	end

	do
		local rawset = rawset

		getmetatable(actionName).__newindex = function(self, key, value)
			rawset(self, key, value)
			UpdateActionNames()
		end
	end
end

--[[----------------------------------------------------------------------------
Initialize
------------------------------------------------------------------------------]]
LibStub('AceConfigRegistry-3.0'):RegisterOptionsTable(addonName .. "Override", overrideOptions, true)

addon.config[1]:AssignOptions(addonName .. "Override")

--addon.config[1]:SetDesc(L["These options allow you to change the behavior of %s."]:format(L["individual actions"]))
addon.config[1]:SetDesc(L["%s  This feature is a Work-in-Progress and as such it's settings may be wiped with newer versions (if any).  It is being provided now in this unfinished state due to the versatility it offers."]:format(L["These options allow you to change the behavior of %s."]:format(L["individual actions"])))

-- Update local settings when the profile changes
local function LoadProfile()
	_settings = addon.db.profile
	_override = _settings.override
	overrideTab = nil
	UpdateActionNames()
end
LoadProfile()

local RegisterCallback = addon.db.RegisterCallback
RegisterCallback(addon.config[1], 'OnProfileChanged', LoadProfile)
RegisterCallback(addon.config[1], 'OnProfileCopied', LoadProfile)
RegisterCallback(addon.config[1], 'OnProfileReset', LoadProfile)
